import astropy.units as u
import pytest
import xarray as xr

from psipy.model import base, mas


def test_mas_model(mas_directory):
    # Check that loading a single file works
    mas_model = mas.MASOutput(mas_directory)
    assert isinstance(mas_model, base.ModelOutput)

    rho = mas_model['rho']
    assert isinstance(rho, base.Variable)
    assert isinstance(rho.data, xr.DataArray)
    assert rho.unit == u.N / u.cm**3



"""
Tools for reading MAS (Magnetohydrodynamics on a sphere) model outputs.
"""
import glob
from pathlib import Path

import numpy as np
import xarray as xr

from .util import read_hdf4, read_hdf5


__all__ = ['read_mas_file', 'get_mas_variables']


def read_mas_file(directory, var):
    """
    Read in a single MAS output file.

    Parameters
    ----------
    directory :
        Directory to look in.
    var : str
        Variable name.

    Returns
    -------
    data : xarray.DataArray
        Loaded data.
    """
    directory = Path(directory)
    files = glob.glob(str(directory / f'{var}[0-9][0-9][0-9].h*'))
    if not len(files):
        raise FileNotFoundError(f'Could not find file for variable "{var}" in '
                                f'directory {directory}')
    f = Path(files[0])
    if f.suffix == '.hdf':
        data, coords = read_hdf4(f)
    elif f.suffix == '.h5':
        data, coords = read_hdf5(f)

    dims = ['phi', 'theta', 'r']
    # Convert from co-latitude to latitude
    coords[1] = np.pi / 2 - np.array(coords[1])
    data = xr.DataArray(data=data, coords=coords, dims=dims)
    return data


def get_mas_variables(path):
    """
    Return a list of variables present in a given directory.

    Parameters
    ----------
    path :
        Path to the folder containing the MAS data files.

    Returns
    -------
    var_names : list
        List of variable names present in the given directory.
    """
    files = glob.glob(str(path / '*[0-9][0-9][0-9].h*'))
    # Get the variable name from the filename
    # Here we take the filename before .hdf, and remove the last three
    # characters which give the timestep
    var_names = [Path(f).stem.split('.h')[0][:-3] for f in files]
    if not len(var_names):
        raise FileNotFoundError(f'No variable files found in {path}')
    return var_names



from matplotlib.projections.polar import ThetaFormatter
import numpy as np


def clear_axes_labels(ax):
    """
    Remove labels from both x and y axes.
    """
    ax.set_xlabel('')
    ax.set_ylabel('')


def set_theta_formatters(ax):
    """
    Set both x and y axes to have theta formatters (ie. degrees)
    """
    for axis in [ax.xaxis, ax.yaxis]:
        axis.set_major_formatter(ThetaFormatter())
        axis.set_minor_formatter(ThetaFormatter())


def setup_polar_ax(ax):
    if ax is None:
        ax = plt.gca(projection='polar')
    if ax.name != 'polar':
        raise ValueError('ax must have a polar projection')
    return ax


def format_polar_ax(ax):
    # Plot formatting
    ax.set_rlim(0)
    ax.set_thetalim(-np.pi / 2, np.pi / 2)
    clear_axes_labels(ax)

    # Tick label formatting
    # Set theta ticks
    ax.set_xticks([])


def format_equatorial_ax(ax):
    # Plot formatting
    ax.set_rlim(0)
    ax.set_thetalim(0, 2 * np.pi)
    clear_axes_labels(ax)

    # Tick label formatting
    # Remove theta ticks
    ax.set_xticks([])



"""
Comparing in-situ data to model output
======================================

In this example, we'll see how to compare in-situ data taken by a spacecraft
to equivalent observables in the model along the 1D spacecraft trajectory.

This consists of three steps:
1. Load the spacecraft data
2. Generate the spacecraft trajectory at the in-situ data timestamps
3. Use this trajectory to take samples in the 3D model output

The first two steps are accomplished using ``heliopy.data``
(to get the in-situ data) and ``heliopy.spice`` (to get the trajectory). This
is then fed into `Variable.sample_at_coords` to get the model values, which
we then compare to the in-situ data.
"""
###############################################################################
# First, load the required modules.
import matplotlib.pyplot as plt

from heliopy.data import psp
import heliopy.data.spice as spicedata
import heliopy.spice as spice

from psipy.model import MASOutput
from psipy.data import sample_data

###############################################################################
# Load a set of MAS output files.
mas_path = sample_data.mas_helio()
model = MASOutput(mas_path)
print(model.variables)

###############################################################################
# Load PSP data.
#
# Here we load the merged magnetic field and plasma data product, which has
# an hourly data cadence.
starttime = '2018-10-22'
endtime = '2018-11-21'
psp_data = psp.merged_mag_plasma(starttime, endtime)
print(psp_data.columns)

###############################################################################
# Generate the PSP trajectory.
#
# We take the timestamps from the previously loaded data, and use `heliopy.spice`
# to generate the trajectory at these times.
times = psp_data.index

spicedata.get_kernel('psp')
spicedata.get_kernel('psp_pred')
psp_traj = spice.Trajectory('SPP')
psp_traj.generate_positions(times, 'Sun', 'IAU_SUN')
psp_coords = psp_traj.coords
print(psp_coords)

###############################################################################
# Take a sample of the radial velocity.
#
# Here we start by getting the radial velocity `Variable` from the model, and
# then use the PSP corodinate information to sample it.
vr_model = model['vr']
vr_sampled = vr_model.sample_at_coords(psp_coords.lon,
                                       psp_coords.lat,
                                       psp_coords.radius)

###############################################################################
# We can now plot a comparison between the model and in-situ measurements.
fig, ax = plt.subplots()
ax.plot(times, vr_sampled, label='Model')
ax.plot(times, psp_data.quantity('VR'), label='PSP')

ax.set_ylabel(r'$v_{r}$ (km/s)')
ax.legend()
fig.autofmt_xdate()

###############################################################################
# To finish, we'll perform the same comparison, but with a few different
# variables.
fig, axs = plt.subplots(nrows=3, sharex=True)
for ax, mas_name, psp_name in zip(axs,
                                  ['rho', 'vr', 'br'],
                                  ['protonDensity', 'VR', 'BR']):
    model_var = model[mas_name]
    sampled = model_var.sample_at_coords(psp_coords.lon,
                                         psp_coords.lat,
                                         psp_coords.radius)

    in_situ = psp_data.quantity(psp_name)

    # Note that we convert the sampled data to the same units as the PSP data
    ax.plot(times, sampled.to(in_situ.unit), label='Model')
    ax.plot(times, in_situ, label='PSP')

    ax.set_ylabel(str(model_var.unit))
    ax.legend()

fig.autofmt_xdate()

plt.show()



"""
Plotting radial slices
======================

This example shows how to plot slices of constant radial distance from a MAS model
output.
"""
###############################################################################
# First, load the required modules.
from psipy.model import MASOutput
from psipy.data import sample_data

import matplotlib.pyplot as plt

###############################################################################
# Next, load a set of MAS output files. You will need to change this line to
# point to a folder with MAS files in them.
mas_path = sample_data.mas_helio()
model = MASOutput(mas_path)

###############################################################################
# Each MAS model contains a number of variables. The variable names can be
# accessed using the ``.variables`` attribute.
print(model.variables)

###############################################################################
# Set parameters for plotting. The first line will give us a horizontal
# errorbar underneath the plots. The second line is the index to select for the
# radial slice.
cbar_kwargs = {'orientation': 'horizontal'}
r_idx = 139


###############################################################################
# Plot the slices
fig, axs = plt.subplots(ncols=3, tight_layout=True)

ax = axs[0]
model['vr'].plot_radial_cut(r_idx, ax=ax, cbar_kwargs=cbar_kwargs)
model['br'].contour_radial_cut(r_idx, levels=[0], ax=ax, colors='white')
ax = axs[1]
model['rho'].plot_radial_cut(r_idx, ax=ax, cbar_kwargs=cbar_kwargs)
model['br'].contour_radial_cut(r_idx, levels=[0], ax=ax, colors='white')
ax = axs[2]
model['br'].plot_radial_cut(r_idx, ax=ax, cbar_kwargs=cbar_kwargs)
model['br'].contour_radial_cut(r_idx, levels=[0], ax=ax, colors='black')

plt.show()



import pytest
import xarray as xr

from psipy.io import mas


def test_read_mas_file(mas_directory):
    # Check that loading a single file works
    data = mas.read_mas_file(mas_directory, 'rho')
    assert isinstance(data, xr.DataArray)



import xarray as xr

from psipy.io import pluto


def test_read_pluto_files(pluto_directory):
    # Check that loading a single file works
    data = pluto.read_pluto_files(pluto_directory, 'rho')
    assert isinstance(data, xr.DataArray)



"""
Plotting constant longitude slices
==================================

This example shows how to plot slices of constant longitude from a MAS model
output.
"""
###############################################################################
# First, load the required modules.
from psipy.model import MASOutput
from psipy.data import sample_data

import matplotlib.pyplot as plt

###############################################################################
# Next, load a set of MAS output files. You will need to change this line to
# point to a folder with MAS files in them.
mas_path = sample_data.mas_helio()
model = MASOutput(mas_path)

###############################################################################
# Each MAS model contains a number of variables. The variable names can be
# accessed using the ``.variables`` attribute.
print(model.variables)

###############################################################################
# Set parameters for plotting. The first line will give us a horizontal
# errorbar underneath the plots. The second line is the index to select for the
# longitude slice.
cbar_kwargs = {'orientation': 'horizontal'}
phi_idx = 40

###############################################################################
# Plot the slices
#
# Note that for density (rho) and pressure (p) we first normalise the data
# relative to a power law decrease, to make it easer to see spatial variations.
fig = plt.figure()
axs = [plt.subplot(1, 2, i + 1, projection='polar') for i in range(2)]

ax = axs[0]
model['vr'].plot_phi_cut(phi_idx, ax=ax, cbar_kwargs=cbar_kwargs)

ax = axs[1]
rho = model['rho']
rho_r2 = rho.radial_normalized(2)
rho_r2.plot_phi_cut(phi_idx, ax=ax, cbar_kwargs=cbar_kwargs)


# Add a contour of br = 0 (the heliopsheric current sheet) to all the axes
for ax in axs:
    model['br'].contour_phi_cut(phi_idx, levels=[0], ax=ax,
                                colors='white', linestyles='--', linewidths=1)

plt.show()



"""
Tools for reading pluto model outputs.
"""
import glob
from pathlib import Path

import numpy as np
import xarray as xr


__all__ = ['read_pluto_files', 'get_pluto_variables', 'read_pluto_grid']


def get_timestep(path):
    """
    Get the timestep from a filename.
    """
    path = Path(path)
    fname = path.stem
    tstep = fname.split('.')[-1]
    return int(tstep)


def read_pluto_files(directory, var):
    """
    Read in a single variable from a set of PLUTO output files.

    Parameters
    ----------
    directory :
        Directory to look in.
    var : str
        Variable name.

    Returns
    -------
    data : xarray.DataArray
        Loaded data.
    """
    directory = Path(directory)
    files = glob.glob(str(directory / f'{var}*.dbl'))
    if not len(files):
        raise FileNotFoundError(f'Could not find any files for variable "{var}" in '
                                f'directory {directory}')
    files.sort()
    all_data = []
    times = []
    for file in files:
        times.append(get_timestep(file))
        data, grid = read_pluto_dbl(file)
        all_data.append(data)

    # Take grid centers as the grid points
    coords = [np.mean(g, axis=1) for g in grid]
    # Convert from co-latitude to latitude
    coords[1] = np.pi / 2 - np.array(coords[1])
    coords = [times] + coords

    dims = ['time', 'phi', 'theta', 'r']
    data = xr.DataArray(data=all_data, coords=coords, dims=dims)
    return data.isel(time=1)


def read_pluto_grid(path):
    """
    Read in a single PLUTO grid file.

    Parameters
    ----------
    path :
        Path to the ``grid.out`` file.

    Returns
    -------
    dim1, dim2, dim3 : numpy.ndarray
        Coordinate values along each dimension. These are (n, 2) shaped arrays,
        with the two columns being the minimum and maximum coordinate values
        of a given cell.
    """
    with open(path, 'r') as f:
        lines = f.readlines()
    n_header_lines = np.sum([line[0] == '#' for line in lines])
    # Read size of dimensions
    n_dim_1 = int(lines[n_header_lines].split('\n')[0])
    n_dim_2 = int(lines[n_header_lines + n_dim_1 + 1].split('\n')[0])
    n_dim_3 = int(lines[n_header_lines + n_dim_1 + n_dim_2 + 2].split('\n')[0])

    # Read in coordinate values
    dim_1 = np.loadtxt(path, skiprows=n_header_lines + 1, max_rows=n_dim_1)
    dim_2 = np.loadtxt(path, skiprows=n_header_lines + n_dim_1 + 2, max_rows=n_dim_2)
    dim_3 = np.loadtxt(path, skiprows=n_header_lines + n_dim_1 + n_dim_2 + 3, max_rows=n_dim_3)
    return dim_1[:, 1:], dim_2[:, 1:], dim_3[:, 1:]


def read_pluto_dbl(path):
    """
    Read in a single PLUTO output file.

    Parameters
    ----------
    path :
        Path to the dbl file.

    Returns
    -------
    data : numpy.ndarray
        3D array of data values
    grid : list
        Each item is a (n, 2) shaped array of the min/max limits of the cells
        in each coordinate.

    Notes
    -----
    There must be a ``grid.out`` file present in the same directory as the file
    being read.
    """
    path = Path(path)
    data = np.fromfile(path, np.float64)
    grid = read_pluto_grid(path.parent / 'grid.out')

    grid = grid[::-1]
    grid_dims = list(g.shape[0] for g in grid)
    data = data.reshape(grid_dims)

    return data, grid


def get_pluto_variables(directory):
    """
    Return a list of variables present in a given directory.

    Parameters
    ----------
    directory :
        Path to the folder containing the PLUTO data files.

    Returns
    -------
    var_names : list
        List of variable names present in the given directory.
    """
    files = glob.glob(str(Path(directory) / '*.dbl'))
    # Get the variable name from the filename
    # Take anything before the . in the first three characters
    var_names = [Path(f).stem[:3].split('.')[0] for f in files]
    # Only return unique names
    var_names = list(set(var_names))
    if not len(var_names):
        raise FileNotFoundError(f'No variable files found in {directory}')
    var_names.sort()
    return var_names



import abc
from pathlib import Path

from .variable import Variable


__all__ = ['ModelOutput']


class ModelOutput(abc.ABC):
    r"""
    The results from a single model run.

    This is a storage object that contains a number of `Variable` objects. It
    is not designed to be used directly, but must be sub-classed for different
    models.

    Data is stored in the ``_data`` attribute. This is a mapping of variable
    names to `xarray.DataArray` variables. Each data array must have

    - Three dimensions
    - These dimensions must be labelled ``['r', 'theta', 'phi']``
    - The phi values must be latitude and *not* co-latitude (ie. must be in
      the range :math:`[-\pi / 2, \pi / 2]`)
    - The theta values must be in the range :math:`[0, 2\pi]`

    Notes
    -----
    Variables are loaded on demand. To see the list of available variables
    use `ModelOutput.variables`, and to see the list of already loaded variables
    use `ModelOutput.loaded_variables`.

    Parameters
    ----------
    path :
        Path to the directry containing the model output files.
    """
    def __init__(self, path):
        self.path = Path(path)
        # Leave data empty for now, as we want to load on demand
        self._data = {}
        self._variables = self.get_variables()
        self._variables.sort()

    def __str__(self):
        return (f'{self.__class__.__name__}\n'
                f'Variables: {self.variables}')

    def __getitem__(self, var):
        if var not in self.variables:
            raise RuntimeError(f'{var} not in list of known variables: '
                               f'{self._variables}')
        if var not in self.loaded_variables:
            self._data[var] = self.load_file(var)

        # Get units
        try:
            unit, factor = self.get_unit(var)
        except Exception as e:
            raise RuntimeError('Do not know what units are for '
                               f'variable "{var}"') from e

        data = self._data[var] * factor
        return Variable(data, var, unit)

    # Abstract methods start here
    #
    # These are methods that must be defined by classes that inherit from this
    # class
    @abc.abstractmethod
    def get_variables(self):
        """
        Returns
        -------
        list :
            A list of all variable names present in the directory.
        """

    @abc.abstractmethod
    def load_file(self, var):
        """
        Load data for variable *var*.
        """

    @abc.abstractmethod
    def get_unit(self, var):
        """
        Return the units for a variable, and the factor needed to convert
        from the model output to those units.

        Returns
        -------
        unit : `astropy.units.Unit`
        factor : float
        """

    # Properties start here
    @property
    def loaded_variables(self):
        """
        List of loaded variable names.
        """
        return list(self._data.keys())

    @property
    def variables(self):
        """
        List of all variable names present in the directory.
        """
        return self._variables
